@@ -11,6 +11,8 @@ from threading import Thread, Event
 from websocket import WebSocketTimeoutException, WebSocketException, WebSocketConnectionClosedException, \
     create_connection
 
+from .errors import CallMethodError
+
 
 class GenericAttr(object):
     def __init__(self, name, tab):
@@ -52,8 +54,8 @@ class ChromiumDriver(object):
         self._ws = None
 
         self._recv_th = Thread(target=self._recv_loop)
-        self._recv_th.daemon = True
         self._handle_event_th = Thread(target=self._handle_event_loop)
+        self._recv_th.daemon = True
         self._handle_event_th.daemon = True
 
         self._stopped = Event()
@@ -99,7 +101,8 @@ class ChromiumDriver(object):
 
                 except Empty:
                     if self.has_alert:
-                        return {'result': {'alert': True}}
+                        return {'error': {'message': 'alert exists'}, 'type': 'alert_exists'}
+
                     if isinstance(timeout, (int, float)) and timeout <= 0:
                         raise TimeoutError(f"调用{message['method']}超时。")
 
@@ -146,9 +149,8 @@ class ChromiumDriver(object):
             if event['method'] in self.event_handlers:
                 try:
                     self.event_handlers[event['method']](**event['params'])
-                except Exception:
-                    pass
-                    # raise RuntimeError(f"回调函数 {event['method']} 错误：{e}")
+                except Exception as e:
+                    raise RuntimeError(f"\n回调函数 {self.event_handlers[event['method']].__name__} 错误：\n{e}")
 
             self.event_queue.task_done()
 
@@ -165,19 +167,23 @@ class ChromiumDriver(object):
         :return: 执行结果
         """
         if not self._started:
-            raise RuntimeError("不能在启动前调用方法。")
+            self.start()
+            # raise RuntimeError("不能在启动前调用方法。")
         if args:
-            raise CallMethodException("参数必须是key=value形式。")
+            raise CallMethodError("参数必须是key=value形式。")
 
         if self._stopped.is_set():
-            return {'tab_closed': True}
+            return {'error': 'tab closed', 'type': 'tab_closed'}
 
         timeout = kwargs.pop("_timeout", None)
         result = self._send({"method": _method, "params": kwargs}, timeout=timeout)
         if result is None:
-            return {'tab_closed': True}
+            return {'error': 'tab closed', 'type': 'tab_closed'}
         if 'result' not in result and 'error' in result:
-            raise CallMethodException(f"\n调用方法：{_method}\n参数：{kwargs}\n错误：{result['error']['message']}")
+            return {'error': result['error']['message'],
+                    'type': result.get('type', 'call_method_error'),
+                    'method': _method,
+                    'args': kwargs}
 
         return result['result']
 
@@ -201,12 +207,13 @@ class ChromiumDriver(object):
         if self._stopped.is_set():
             return False
         if not self._started:
-            raise RuntimeError("Driver正在运行。")
+            return True
 
         self.status = self._STOPPED_
         self._stopped.set()
         if self._ws:
             self._ws.close()
+            self._ws = None
         self.event_handlers.clear()
         self.method_results.clear()
         self.event_queue.queue.clear()
@@ -237,7 +244,3 @@ class ChromiumDriver(object):
         return f"<ChromiumDriver {self.id}>"
 
     __repr__ = __str__
-
-
-class CallMethodException(Exception):
-    pass
