@@ -11,13 +11,13 @@ from warnings import warn
 
 from requests import Session
 
-from .chromium_base import ChromiumBase, Timeout
-from .chromium_driver import ChromiumDriver, CallMethodException
+from .chromium_base import ChromiumBase, Timeout, ChromiumBaseSetter, ChromiumBaseWaiter
+from .chromium_driver import ChromiumDriver
 from .chromium_tab import ChromiumTab
+from .commons.browser import connect_browser
+from .commons.web import set_session_cookies
 from .configs.chromium_options import ChromiumOptions
-from .configs.driver_options import DriverOptions
-from .functions.browser import connect_browser
-from .functions.web import set_session_cookies
+from .errors import CallMethodError, BrowserConnectError
 from .session_page import DownloadSetter
 
 
@@ -34,78 +34,85 @@ class ChromiumPage(ChromiumBase):
         self._download_path = None
         super().__init__(addr_driver_opts, tab_id, timeout)
 
-    def _connect_browser(self, addr_driver_opts=None, tab_id=None):
-        """连接浏览器，在第一次时运行
-        :param addr_driver_opts: 浏览器地址、ChromiumDriver对象或DriverOptions对象
-        :param tab_id: 要控制的标签页id，不指定默认为激活的
+    def _set_start_options(self, addr_driver_opts, none):
+        """设置浏览器启动属性
+        :param addr_driver_opts: 'ip:port'、ChromiumDriver、ChromiumOptions
+        :param none: 用于后代继承
         :return: None
         """
-        # 接管或启动浏览器
-        self._chromium_init()
-        if addr_driver_opts is None or isinstance(addr_driver_opts, (ChromiumOptions, DriverOptions)):
-            self._driver_options = addr_driver_opts or ChromiumOptions()  # 从ini文件读取
-            self.address = self._driver_options.debugger_address
-            self.process = connect_browser(self._driver_options)[1]
-            json = self._control_session.get(f'http://{self.address}/json').json()
-            tab_id = [i['id'] for i in json if i['type'] == 'page'][0]
+        if not addr_driver_opts or str(type(addr_driver_opts)).endswith(("ChromiumOptions'>", "DriverOptions'>")):
+            self._driver_options = addr_driver_opts or ChromiumOptions(addr_driver_opts)
 
         # 接收浏览器地址和端口
         elif isinstance(addr_driver_opts, str):
-            self.address = addr_driver_opts
             self._driver_options = ChromiumOptions()
             self._driver_options.debugger_address = addr_driver_opts
-            self.process = connect_browser(self._driver_options)[1]
-            if not tab_id:
-                json = self._control_session.get(f'http://{self.address}/json').json()
-                tab_id = [i['id'] for i in json if i['type'] == 'page'][0]
 
         # 接收传递过来的ChromiumDriver，浏览器
         elif isinstance(addr_driver_opts, ChromiumDriver):
-            self._tab_obj = addr_driver_opts
-            self.address = addr_driver_opts.address
-            self.process = None
             self._driver_options = ChromiumOptions(read_file=False)
             self._driver_options.debugger_address = addr_driver_opts.address
+            self._tab_obj = addr_driver_opts
 
         else:
             raise TypeError('只能接收ChromiumDriver或ChromiumOptions类型参数。')
 
-        self._set_options()
-        self._set_chromium_options()
-        self._init_page(tab_id)
-        self._get_document()
-        self._first_run = False
+        self.address = self._driver_options.debugger_address
 
-    def _set_options(self):
-        """设置WebPage中与s模式共用的配置，便于WebPage覆盖掉"""
+    def _set_runtime_settings(self):
+        """设置运行时用到的属性"""
         self._timeouts = Timeout(self,
                                  page_load=self._driver_options.timeouts['pageLoad'],
                                  script=self._driver_options.timeouts['script'],
                                  implicit=self._driver_options.timeouts['implicit'])
         self._page_load_strategy = self._driver_options.page_load_strategy
+        self._download_path = self._driver_options.download_path
 
-    def _set_chromium_options(self):
-        """设置浏览器专有的配置"""
-        self._alert = Alert()
-        self._window_setter = None
-
-    def _init_page(self, tab_id=None):
-        """新建页面、页面刷新、切换标签页后要进行的cdp参数初始化
-        :param tab_id: 要跳转到的标签页id
+    def _connect_browser(self, tab_id=None):
+        """连接浏览器，在第一次时运行
+        :param tab_id: 要控制的标签页id，不指定默认为激活的
         :return: None
         """
-        super()._init_page(tab_id)
+        self._chromium_init()
+
+        if not self._tab_obj:  # 不是传入driver的情况
+            connect_browser(self._driver_options)
+            if not tab_id:
+                json = self._control_session.get(f'http://{self.address}/json').json()
+                tab_id = [i['id'] for i in json if i['type'] == 'page']
+                if not tab_id:
+                    raise BrowserConnectError('浏览器连接失败，可能是浏览器版本原因。')
+                tab_id = tab_id[0]
+
+            self._driver_init(tab_id)
+
+        self._page_init()
+        self._get_document()
+        self._first_run = False
+
+    def _page_init(self):
+        """页面相关设置"""
         ws = self._control_session.get(f'http://{self.address}/json/version').json()['webSocketDebuggerUrl']
         self._browser_driver = ChromiumDriver(ws.split('/')[-1], 'browser', self.address)
         self._browser_driver.start()
+
+        self._alert = Alert()
         self._tab_obj.Page.javascriptDialogOpening = self._on_alert_open
         self._tab_obj.Page.javascriptDialogClosed = self._on_alert_close
+
+        self._rect = None
         self._main_tab = self.tab_id
         try:
-            self.download_set.by_DownloadKit()
-        except RuntimeError:
+            self.download_set.by_browser()
+        except CallMethodError:
             pass
 
+        self._process_id = None
+        for i in self.browser_driver.SystemInfo.getProcessInfo()['processInfo']:
+            if i['type'] == 'browser':
+                self._process_id = i['id']
+                break
+
     @property
     def browser_driver(self):
         """返回用于控制浏览器cdp的driver"""
@@ -134,17 +141,14 @@ class ChromiumPage(ChromiumBase):
     @property
     def process_id(self):
         """返回浏览器进程id"""
-        try:
-            return self._driver.SystemInfo.getProcessInfo()['id']
-        except Exception:
-            return None
+        return self._process_id
 
     @property
-    def set_window(self):
-        """返回用于设置窗口大小的对象"""
-        if self._window_setter is None:
-            self._window_setter = WindowSetter(self)
-        return self._window_setter
+    def set(self):
+        """返回用于等待的对象"""
+        if self._set is None:
+            self._set = ChromiumPageSetter(self)
+        return self._set
 
     @property
     def download_path(self):
@@ -164,6 +168,19 @@ class ChromiumPage(ChromiumBase):
         """返回下载器对象"""
         return self.download_set._switched_DownloadKit
 
+    @property
+    def rect(self):
+        if self._rect is None:
+            self._rect = ChromiumTabRect(self)
+        return self._rect
+
+    @property
+    def wait(self):
+        """返回用于等待的对象"""
+        if self._wait is None:
+            self._wait = ChromiumPageWaiter(self)
+        return self._wait
+
     def get_tab(self, tab_id=None):
         """获取一个标签页对象
         :param tab_id: 要获取的标签页id，为None时获取当前tab
@@ -172,62 +189,6 @@ class ChromiumPage(ChromiumBase):
         tab_id = tab_id or self.tab_id
         return ChromiumTab(self, tab_id)
 
-    def get_screenshot(self, path=None, as_bytes=None, full_page=False, left_top=None, right_bottom=None):
-        """对页面进行截图，可对整个网页、可见网页、指定范围截图。对可视范围外截图需要90以上版本浏览器支持
-        :param path: 完整路径，后缀可选 'jpg','jpeg','png','webp'
-        :param as_bytes: 是否已字节形式返回图片，可选 'jpg','jpeg','png','webp'，生效时path参数无效
-        :param full_page: 是否整页截图，为True截取整个网页，为False截取可视窗口
-        :param left_top: 截取范围左上角坐标
-        :param right_bottom: 截取范围右下角角坐标
-        :return: 图片完整路径或字节文本
-        """
-        if as_bytes:
-            if as_bytes is True:
-                pic_type = 'png'
-            else:
-                if as_bytes not in ('jpg', 'jpeg', 'png', 'webp'):
-                    raise ValueError("只能接收'jpg', 'jpeg', 'png', 'webp'四种格式。")
-                pic_type = 'jpeg' if as_bytes == 'jpg' else as_bytes
-
-        else:
-            if not path:
-                raise ValueError('保存为文件时必须传入路径。')
-            path = Path(path)
-            pic_type = path.suffix.lower()
-            if pic_type not in ('.jpg', '.jpeg', '.png', '.webp'):
-                raise TypeError(f'不支持的文件格式：{pic_type}。')
-            pic_type = 'jpeg' if pic_type == '.jpg' else pic_type[1:]
-
-        width, height = self.size
-        if full_page:
-            vp = {'x': 0, 'y': 0, 'width': width, 'height': height, 'scale': 1}
-            png = self._wait_driver.Page.captureScreenshot(format=pic_type, captureBeyondViewport=True, clip=vp)['data']
-        else:
-            if left_top and right_bottom:
-                x, y = left_top
-                w = right_bottom[0] - x
-                h = right_bottom[1] - y
-                vp = {'x': x, 'y': y, 'width': w, 'height': h, 'scale': 1}
-                png = self._wait_driver.Page.captureScreenshot(format=pic_type, captureBeyondViewport=True, clip=vp)[
-                    'data']
-            else:
-                png = self._wait_driver.Page.captureScreenshot(format=pic_type)['data']
-
-        from base64 import b64decode
-        png = b64decode(png)
-
-        if as_bytes:
-            return png
-
-        path.parent.mkdir(parents=True, exist_ok=True)
-        with open(path, 'wb') as f:
-            f.write(png)
-        return str(path.absolute())
-
-    def to_front(self):
-        """激活当前标签页使其处于最前面"""
-        self._control_session.get(f'http://{self.address}/json/activate/{self.tab_id}')
-
     def new_tab(self, url=None, switch_to=True):
         """新建一个标签页,该标签页在最后面
         :param url: 新标签页跳转到的网址
@@ -255,37 +216,35 @@ class ChromiumPage(ChromiumBase):
         else:
             self.run_cdp('Target.createTarget', url='')
 
-    def set_main_tab(self, tab_id=None):
-        """设置主tab
-        :param tab_id: 标签页id，不传入则设置当前tab
-        :return: None
-        """
-        self._main_tab = tab_id or self.tab_id
-
     def to_main_tab(self):
         """跳转到主标签页"""
         self.to_tab(self._main_tab)
 
-    def to_tab(self, tab_id=None, activate=True):
+    def to_tab(self, tab_or_id=None, activate=True):
         """跳转到标签页
-        :param tab_id: 标签页id字符串，默认跳转到main_tab
+        :param tab_or_id: 标签页对象或id，默认跳转到main_tab
         :param activate: 切换后是否变为活动状态
         :return: None
         """
-        self._to_tab(tab_id, activate)
+        self._to_tab(tab_or_id, activate)
 
-    def _to_tab(self, tab_id=None, activate=True, read_doc=True):
+    def _to_tab(self, tab_or_id=None, activate=True, read_doc=True):
         """跳转到标签页
-        :param tab_id: 标签页id字符串，默认跳转到main_tab
+        :param tab_or_id: 标签页对象或id，默认跳转到main_tab
         :param activate: 切换后是否变为活动状态
         :param read_doc: 切换后是否读取文档
         :return: None
         """
         tabs = self.tabs
-        if not tab_id:
+        if not tab_or_id:
             tab_id = self._main_tab
+        elif isinstance(tab_or_id, ChromiumTab):
+            tab_id = tab_or_id.tab_id
+        else:
+            tab_id = tab_or_id
+
         if tab_id not in tabs:
-            tab_id = tabs[0]
+            tab_id = self.latest_tab
 
         if activate:
             self._control_session.get(f'http://{self.address}/json/activate/{tab_id}')
@@ -293,31 +252,28 @@ class ChromiumPage(ChromiumBase):
         if tab_id == self.tab_id:
             return
 
-        self._driver.stop()
-        self._init_page(tab_id)
+        self.driver.stop()
+        self._driver_init(tab_id)
         if read_doc and self.ready_state == 'complete':
             self._get_document()
 
-    def wait_download_begin(self, timeout=None):
-        """等待浏览器下载开始
-        :param timeout: 等待超时时间，为None则使用页面对象timeout属性
-        :return: 是否等到下载开始
-        """
-        return self.download_set.wait_download_begin(timeout)
-
-    def close_tabs(self, tab_ids=None, others=False):
+    def close_tabs(self, tabs_or_ids=None, others=False):
         """关闭传入的标签页，默认关闭当前页。可传入多个
-        :param tab_ids: 要关闭的标签页id，可传入id组成的列表或元组，为None时关闭当前页
+        :param tabs_or_ids: 要关闭的标签页对象或id，可传入列表或元组，为None时关闭当前页
         :param others: 是否关闭指定标签页之外的
         :return: None
         """
         all_tabs = set(self.tabs)
-        if isinstance(tab_ids, str):
-            tabs = {tab_ids}
-        elif tab_ids is None:
+        if isinstance(tabs_or_ids, str):
+            tabs = {tabs_or_ids}
+        elif isinstance(tabs_or_ids, ChromiumTab):
+            tabs = {tabs_or_ids.tab_id}
+        elif tabs_or_ids is None:
             tabs = {self.tab_id}
+        elif isinstance(tabs_or_ids, (list, tuple)):
+            tabs = set(i.tab_id if isinstance(i, ChromiumTab) else i for i in tabs_or_ids)
         else:
-            tabs = set(tab_ids)
+            raise TypeError('tabs_or_ids参数只能传入标签页对象或id。')
 
         if others:
             tabs = all_tabs - tabs
@@ -328,7 +284,7 @@ class ChromiumPage(ChromiumBase):
             return
 
         if self.tab_id in tabs:
-            self._driver.stop()
+            self.driver.stop()
 
         for tab in tabs:
             self._control_session.get(f'http://{self.address}/json/close/{tab}')
@@ -340,12 +296,12 @@ class ChromiumPage(ChromiumBase):
 
         self.to_tab()
 
-    def close_other_tabs(self, tab_ids=None):
+    def close_other_tabs(self, tabs_or_ids=None):
         """关闭传入的标签页以外标签页，默认保留当前页。可传入多个
-        :param tab_ids: 要保留的标签页id，可传入id组成的列表或元组，为None时保存当前页
+        :param tabs_or_ids: 要保留的标签页对象或id，可传入列表或元组，为None时保存当前页
         :return: None
         """
-        self.close_tabs(tab_ids, True)
+        self.close_tabs(tabs_or_ids, True)
 
     def handle_alert(self, accept=True, send=None, timeout=None):
         """处理提示框，可以自动等待提示框出现
@@ -364,19 +320,11 @@ class ChromiumPage(ChromiumBase):
 
         res_text = self._alert.text
         if self._alert.type == 'prompt':
-            self._driver.Page.handleJavaScriptDialog(accept=accept, promptText=send)
+            self.driver.Page.handleJavaScriptDialog(accept=accept, promptText=send)
         else:
-            self._driver.Page.handleJavaScriptDialog(accept=accept)
+            self.driver.Page.handleJavaScriptDialog(accept=accept)
         return res_text
 
-    def hide_browser(self):
-        """隐藏浏览器窗口，只在Windows系统可用"""
-        show_or_hide_browser(self, hide=True)
-
-    def show_browser(self):
-        """显示浏览器窗口，只在Windows系统可用"""
-        show_or_hide_browser(self, hide=False)
-
     def quit(self):
         """关闭浏览器"""
         self._tab_obj.Browser.close()
@@ -402,6 +350,119 @@ class ChromiumPage(ChromiumBase):
         self._alert.response_text = None
         self._tab_obj.has_alert = True
 
+    # --------------准备弃用-------------
+    def set_main_tab(self, tab_id=None):
+        """设置主tab
+        :param tab_id: 标签页id，不传入则设置当前tab
+        :return: None
+        """
+        warn("set_main_tab()方法即将弃用，请用set.main_tab()方法代替。", DeprecationWarning)
+        self.set.main_tab(tab_id)
+
+    @property
+    def set_window(self):
+        """返回用于设置窗口大小的对象"""
+        warn("set_window()方法即将弃用，请用set.window.xxxx()方法代替。", DeprecationWarning)
+        return WindowSetter(self)
+
+    def wait_download_begin(self, timeout=None):
+        """等待浏览器下载开始
+        :param timeout: 等待超时时间，为None则使用页面对象timeout属性
+        :return: 是否等到下载开始
+        """
+        warn("wait_download_begin()方法即将弃用，请用wait.download_begin()方法代替。", DeprecationWarning)
+        return self.download_set.wait_download_begin(timeout)
+
+    def hide_browser(self):
+        """隐藏浏览器窗口，只在Windows系统可用"""
+        warn("hide_browser()方法即将弃用，请用set.hide()方法代替。", DeprecationWarning)
+        show_or_hide_browser(self, hide=True)
+
+    def show_browser(self):
+        """显示浏览器窗口，只在Windows系统可用"""
+        warn("show_browser()方法即将弃用，请用set.show()方法代替。", DeprecationWarning)
+        show_or_hide_browser(self, hide=False)
+
+    def to_front(self):
+        """激活当前标签页使其处于最前面"""
+        warn("to_front()方法即将弃用，请用set.tab_to_front()方法代替。", DeprecationWarning)
+        self.set.tab_to_front()
+
+
+class ChromiumPageWaiter(ChromiumBaseWaiter):
+    def download_begin(self, timeout=None):
+        """等待浏览器下载开始
+        :param timeout: 等待超时时间，为None则使用页面对象timeout属性
+        :return: 是否等到下载开始
+        """
+        return self._driver.download_set.wait_download_begin(timeout)
+
+
+class ChromiumTabRect(object):
+    def __init__(self, page):
+        self._page = page
+
+    @property
+    def browser_location(self):
+        """返回浏览器在屏幕上的坐标，左上角为(0, 0)"""
+        r = self._get_browser_rect()
+        if r['windowState'] in ('maximized', 'fullscreen'):
+            return 0, 0
+        return r['left'] + 7, r['top']
+
+    @property
+    def page_location(self):
+        """返回页面左上角在屏幕中坐标，左上角为(0, 0)"""
+        w, h = self.viewport_location
+        r = self._get_page_rect()['layoutViewport']
+        return w - r['pageX'], h - r['pageY']
+
+    @property
+    def viewport_location(self):
+        """返回视口在屏幕中坐标，左上角为(0, 0)"""
+        w_bl, h_bl = self.browser_location
+        w_bs, h_bs = self.browser_size
+        w_vs, h_vs = self.viewport_size_with_scrollbar
+        return w_bl + w_bs - w_vs, h_bl + h_bs - h_vs
+
+    @property
+    def browser_size(self):
+        """返回浏览器大小"""
+        r = self._get_browser_rect()
+        if r['windowState'] == 'fullscreen':
+            return r['width'], r['height']
+        elif r['windowState'] == 'maximized':
+            return r['width'] - 16, r['height'] - 16
+        else:
+            return r['width'] - 16, r['height'] - 7
+
+    @property
+    def page_size(self):
+        """返回页面总宽高，格式：(宽, 高)"""
+        r = self._get_page_rect()['contentSize']
+        return r['width'], r['height']
+
+    @property
+    def viewport_size(self):
+        """返回视口宽高，不包括滚动条，格式：(宽, 高)"""
+        r = self._get_page_rect()['visualViewport']
+        return r['clientWidth'], r['clientHeight']
+
+    @property
+    def viewport_size_with_scrollbar(self):
+        """返回视口宽高，包括滚动条，格式：(宽, 高)"""
+        r = self._page.run_js('return window.innerWidth.toString() + " " + window.innerHeight.toString();')
+        w, h = r.split(' ')
+        return int(w), int(h)
+
+    def _get_page_rect(self):
+        """获取页面范围信息"""
+        return self._page.run_cdp_loaded('Page.getLayoutMetrics')
+
+    def _get_browser_rect(self):
+        """获取浏览器范围信息"""
+        return self._page.browser_driver.Browser.getWindowForTarget(targetId=self._page.tab_id)['bounds']
+
 
 class ChromiumDownloadSetter(DownloadSetter):
     """用于设置下载参数的类"""
@@ -443,9 +504,9 @@ class ChromiumDownloadSetter(DownloadSetter):
         try:
             self._page.browser_driver.Browser.setDownloadBehavior(behavior='allow', downloadPath=path,
                                                                   eventsEnabled=True)
-        except CallMethodException:
+        except CallMethodError:
             warn('\n您的浏览器版本太低，用新标签页下载文件可能崩溃，建议升级。')
-            self._page.run_cdp('Page.setDownloadBehavior', behavior='allow', downloadPath=path, not_change=True)
+            self._page.run_cdp('Page.setDownloadBehavior', behavior='allow', downloadPath=path)
 
         self.DownloadKit.goal_path = path
 
@@ -455,7 +516,7 @@ class ChromiumDownloadSetter(DownloadSetter):
             self._page.browser_driver.Browser.setDownloadBehavior(behavior='allow', eventsEnabled=True,
                                                                   downloadPath=self._page.download_path)
             self._page.browser_driver.Browser.downloadWillBegin = self._download_by_browser
-        except CallMethodException:
+        except CallMethodError:
             self._page.driver.Page.setDownloadBehavior(behavior='allow', downloadPath=self._page.download_path)
             self._page.driver.Page.downloadWillBegin = self._download_by_browser
 
@@ -466,7 +527,7 @@ class ChromiumDownloadSetter(DownloadSetter):
         try:
             self._page.browser_driver.Browser.setDownloadBehavior(behavior='deny', eventsEnabled=True)
             self._page.browser_driver.Browser.downloadWillBegin = self._download_by_DownloadKit
-        except CallMethodException:
+        except CallMethodError:
             raise RuntimeError('您的浏览器版本太低，不支持此方法，请升级。')
         self._behavior = 'deny'
 
@@ -490,18 +551,27 @@ class ChromiumDownloadSetter(DownloadSetter):
 
     def _cookies_to_session(self):
         """把driver对象的cookies复制到session对象"""
-        ua = self._page.driver.Runtime.evaluate(expression='navigator.userAgent;')['result']['value']
+        ua = self._page.run_cdp('Runtime.evaluate', expression='navigator.userAgent;')['result']['value']
         self.session.headers.update({"User-Agent": ua})
         set_session_cookies(self.session, self._page.get_cookies(as_dict=True))
 
     def _download_by_DownloadKit(self, **kwargs):
         """拦截浏览器下载并用downloadKit下载"""
-        self._page.browser_driver.Browser.cancelDownload(guid=kwargs['guid'])
-        self._page.download.add(file_url=kwargs['url'], goal_path=self._page.download_path,
-                                rename=kwargs['suggestedFilename'])
-        if self._download_th is None or not self._download_th.is_alive():
-            self._download_th = Thread(target=self._wait_download_complete, daemon=False)
-            self._download_th.start()
+        url = kwargs['url']
+        if url.startswith('blob:'):
+            self._page.browser_driver.Browser.setDownloadBehavior(behavior='allow', eventsEnabled=True,
+                                                                  downloadPath=self._page.download_path)
+            sleep(2)
+            self._page.browser_driver.Browser.setDownloadBehavior(behavior='deny', eventsEnabled=True)
+
+        else:
+            self._page.browser_driver.Browser.cancelDownload(guid=kwargs['guid'])
+            self._page.download.add(file_url=url, goal_path=self._page.download_path,
+                                    rename=kwargs['suggestedFilename'])
+            if self._download_th is None or not self._download_th.is_alive():
+                self._download_th = Thread(target=self._wait_download_complete, daemon=False)
+                self._download_th.start()
+
         if self._waiting_download:
             self._download_begin = True
 
@@ -531,8 +601,11 @@ class WindowSetter(object):
     """用于设置窗口大小的类"""
 
     def __init__(self, page):
-        self.driver = page.driver
-        self.window_id = self._get_info()['windowId']
+        """
+        :param page: 页面对象
+        """
+        self._page = page
+        self._window_id = self._get_info()['windowId']
 
     def maximized(self):
         """窗口最大化"""
@@ -573,8 +646,8 @@ class WindowSetter(object):
             if s != 'normal':
                 self._perform({'windowState': 'normal'})
             info = self._get_info()['bounds']
-            width = width or info['width']
-            height = height or info['height']
+            width = width - 16 if width else info['width']
+            height = height + 7 if height else info['height']
             self._perform({'width': width, 'height': height})
 
     def location(self, x=None, y=None):
@@ -588,18 +661,51 @@ class WindowSetter(object):
             info = self._get_info()['bounds']
             x = x if x is not None else info['left']
             y = y if y is not None else info['top']
-            self._perform({'left': x, 'top': y})
+            self._perform({'left': x - 8, 'top': y})
+
+    def hide(self):
+        """隐藏浏览器窗口，只在Windows系统可用"""
+        show_or_hide_browser(self._page, hide=True)
+
+    def show(self):
+        """显示浏览器窗口，只在Windows系统可用"""
+        show_or_hide_browser(self._page, hide=False)
 
     def _get_info(self):
         """获取窗口位置及大小信息"""
-        return self.driver.Browser.getWindowForTarget()
+        return self._page.run_cdp('Browser.getWindowForTarget')
 
     def _perform(self, bounds):
         """执行改变窗口大小操作
         :param bounds: 控制数据
         :return: None
         """
-        self.driver.Browser.setWindowBounds(windowId=self.window_id, bounds=bounds)
+        self._page.run_cdp('Browser.setWindowBounds', windowId=self._window_id, bounds=bounds)
+
+
+class ChromiumPageSetter(ChromiumBaseSetter):
+    def main_tab(self, tab_id=None):
+        """设置主tab
+        :param tab_id: 标签页id，不传入则设置当前tab
+        :return: None
+        """
+        self._page._main_tab = tab_id or self._page.tab_id
+
+    @property
+    def window(self):
+        """返回用于设置浏览器窗口的对象"""
+        return WindowSetter(self._page)
+
+    def tab_to_front(self, tab_or_id=None):
+        """激活标签页使其处于最前面
+        :param tab_or_id: 标签页对象或id，为None表示当前标签页
+        :return: None
+        """
+        if not tab_or_id:
+            tab_or_id = self._page.tab_id
+        elif isinstance(tab_or_id, ChromiumTab):
+            tab_or_id = tab_or_id.tab_id
+        self._page._control_session.get(f'http://{self._page.address}/json/activate/{tab_or_id}')
 
 
 def show_or_hide_browser(page, hide=True):
@@ -620,7 +726,7 @@ def show_or_hide_browser(page, hide=True):
     except ImportError:
         raise ImportError('请先安装：pip install pypiwin32')
 
-    pid = page.process_id or get_browser_progress_id(page.process, page.address)
+    pid = page.process_id
     if not pid:
         return None
     hds = get_chrome_hwnds_from_pid(pid, page.title)
